/*

 Copyright (c) 2001 The Regents of the University of California.
 All rights reserved.
 Permission is hereby granted, without written agreement and without
 license or royalty fees, to use, copy, modify, and distribute this
 software and its documentation for any purpose, provided that the above
 copyright notice and the following two paragraphs appear in all copies
 of this software.

 IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
 THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 ENHANCEMENTS, OR MODIFICATIONS.

 */

package htlc;

import htlc.analysis.DepthFirstAdapter;
import htlc.node.AFormalPort;
import htlc.node.AFormalPorts;
import htlc.node.AStatePort;
import htlc.node.ATaskDeclaration;
import htlc.node.NodeCast;
import htlc.node.Token;

import java.util.ArrayList;
import java.util.Map;


/**
 * 
 * This class implements a scoped symbol table.
 * 
 * @author Christoph Kirsch, cm@eecs.berkeley.edu
 *         Daniel Iercan, daniel.iercan@cs.uni-salzburg.at 2005
 * 
 */
public class ScopedSymbolTable extends DepthFirstAdapter {

	public final Map declarations = new TypedTreeMap(StringComparator.instance,
			StringCast.instance, NodeCast.instance);

	public final ArrayList inputFormalPorts = new ArrayList();
	
	public final ArrayList outputFormalPorts = new ArrayList();
	
	public final ArrayList stateFormalPorts = new ArrayList();
	
	public boolean inInputList = false;
	
	public boolean inOutputList = false;

	public ScopedSymbolTable(SymbolTable symbolTable) {
	}

	// Action code
	
	public void inAFormalPorts(AFormalPorts node){
		ATaskDeclaration parent=(ATaskDeclaration)node.parent();
		inInputList=(node==parent.getInputFormalPorts());
		inOutputList=(node==parent.getOutputFormalPorts());
	}
	
	public void outAFormalPorts(AFormalPorts node){
		inInputList=inOutputList=false;
	}

	public void outAFormalPort(AFormalPort node) {
		final String name = node.getPortName().getText();

//		if (symbolTable.declarations.containsKey(name)) {
//			errorRedefined(node.getPortName(), name);
//		}

		if (declarations.put(name, node) != null) {
			errorMultipleUse(node.getPortName(), name);
		}
		
		if(inInputList){
			inputFormalPorts.add(node);
		}
		
		if(inOutputList){
			outputFormalPorts.add(node);
		}
	}
	
	public void outAStatePort(AStatePort node) {
		final String name = node.getStateName().getText();

//		if (symbolTable.declarations.containsKey(name)) {
//			errorRedefined(node.getPortName(), name);
//		}

		if (declarations.put(name, node) != null) {
			errorMultipleUse(node.getStateName(), name);
		}
		
		stateFormalPorts.add(node);
	}
	
	//misc code

	private static void errorMultipleUse(Token token, String name) {
		throw new RuntimeException("[" + token.getLine() + "," + token.getPos()
				+ "] " + "Multiple use of " + name + ".");
	}

	public String toString() {
		final StringBuffer s = new StringBuffer();
		final String nl = System.getProperty("line.separator");

		s.append("Declarations:");
		s.append(nl);
		s.append(declarations);
		s.append(nl);

		return s.toString();
	}
}
